/*
 */

#ifndef _COMMON_H_
#define _COMMON_H_

#include "opcodes.h"

typedef union uGPR	tGPR;
typedef struct sSReg	tSReg;
typedef struct sDesc	tDesc;
typedef struct sRFlags	tRFlags;
typedef struct sCR0	tCR0;
typedef struct sCR4	tCR4;
typedef struct sCPU	tCPU;

union uGPR
{
	struct {
		Uint8	L, H;
	};
	Uint16	X;
	Uint32	E;
	Uint64	R;
};

struct sSReg
{
	Uint16	Sel;
	Uint64	Base;
	Uint64	Limit;
	struct
	{
		 int	DPL;
		BOOL	Code;
		BOOL	RW;
		BOOL	DC;	// Direction / Conforming
		BOOL	Size;	// 16/32 bit
	}	Flags;
	char	*Name;
};

struct sDesc
{
	Uint16	Limit;
	Uint64	Base;
};

struct sRFlags
{
	BOOL	C;	// Carry
	// 1
	BOOL	P;	// Pairity
	// 0
	BOOL	A;	// Auxilary Flag
	// 0
	BOOL	Z;	// Zero
	BOOL	S;	// Sign
	BOOL	T;	// Trap
	BOOL	I;	// Interrupt Enable Flag
	BOOL	D;	// Direction Flag
	BOOL	O;	// Overflow
	 int	IOPL;	// 2-bit Permission Level
	BOOL	NT;	// Nested Task
	// 0
	
	BOOL	R;	// Resume Flag
	BOOL	VM;	// VM8086 Mode
	BOOL	AC;	// Alignment Check
	BOOL	VI;	// Virtual Interrupt Flag
	BOOL	VIP;	// Virtual Interrupt Pending
	BOOL	ID;	// ID Flag
	// 0 til 31
};

struct sCR0
{
	BOOL	PE;	// Protected Mode Enabled
	BOOL	MP;	// Monitor Coprocessor
	BOOL	EM;	// Emulation (CPU does not have a coprocessor)
	BOOL	TS;	// Task Switched
	BOOL	ET;	// Extension Type
	BOOL	NE;	// Numeric Error - 
	// 6-15 zero
	BOOL	WP;	// Write Protect (Causes a page fault if a supervisor writes to a RO user page)
	// 0
	BOOL	AM;	// Alignment Mask (Enables use of EFLAGS.AC)
	// 19-28 zero
	BOOL	NW;	// Not-Write Through
	BOOL	CD;	// Cache Disable
	BOOL	PG;	// Paging
};

struct sCR4
{
	BOOL	VME;
	BOOL	PVI;
	BOOL	TSD;
	BOOL	DE;
	BOOL	PSE;
	BOOL	PAE;
	BOOL	MCE;
	BOOL	PGE;
	BOOL	PCE;
	BOOL	OSFXSR;
	BOOL	OSXMMEXCPT;
	// 0
	// 0
	BOOL	VMXE;
	BOOL	SMXE;
	// 0
	BOOL	OSXSAVE;
};

/*
struct sModRM
{
	unsigned Mod:	2;
	unsigned Reg:	3;
	unsigned RM:	3;
};

struct sSIB
{
	unsigned Scale:	2;
	unsigned Index:	3;
	unsigned Base:	3;
};

struct sREX
{
	unsigned Ident:	4;	// 0x4
	unsigned W:	1;	// Word Bit
	unsigned R:	1;	// Extension of ModR/M Reg field
	unsigned X:	1;	// Extension of SIB index field
	unsigned B:	1;	// General Extension bit
};
*/

enum eCPU_Modes
{
	MODE_OFF,
	MODE_LM,
	MODE_COMPAT,
	MODE_PM,
	MODE_VM8086,
	MODE_RM
};

enum eGPRs
{
	REG_A, REG_C, REG_D, REG_B,
	REG_SP, REG_BP, REG_SI, REG_DI
};

struct sCPU
{
	//! General Registers
	// A, C, D, B, SP, BP, SI, DI
	// R8, ..., R15
	tGPR	GPR[16];
	
	//! Segment Registers
	tSReg	CS, SS;
	tSReg	DS, ES;
	tSReg	FS, GS;
	
	Uint64	RIP;
	tRFlags	RFlags;
	
	tDesc	GDTR;
	tSReg	LDTR;
	tDesc	IDTR;
	
	tCR0	CR0;
	Uint64	CR2;
	Uint64	CR3;
	tCR4	CR4;
	
	 int	Num;
	Uint	Timestamp;
	 int	Mode;
	
	struct
	{
		BOOL	Halted;
	}	StateFlags;
	
	struct
	{
		BOOL	Trace;
		BOOL	DumpRegs;
	} Debug;
	
	
	// -- Decoding State
	tSReg	*Seg;	// Current Data Segment
	tModRM	ModRM;
	tSIB	SIB;
	tREX	REX;
	
	 int	AtomicAccess;
	 int	OvrAddress;
	 int	OvrOperand;
	 
	 int	NestedExceptions;
};

// === FUNCTIONS ===
// --- Misc ---
extern Uint32	GetFlags(tCPU *Regs);
extern void	SetFlags(tCPU *Regs, Uint64 Value, int Bits);
extern Uint32	GetCR0(tCPU *Regs);
extern void	SetCR0(tCPU *Regs, Uint32 Value);
extern Uint32	GetCR4(tCPU *Regs);
extern void	SetCR4(tCPU *Regs, Uint32 Value);
// --- Descriptors ---
extern void	SetSegment( tCPU *Regs, tSReg *Seg, Uint16 Selector );
extern void	SetGDT( tCPU *Regs, Uint64 Addr );
extern void	SetIDT( tCPU *Regs, Uint64 Addr );
// --- Memory ---
extern void	ReadLinear( tCPU *Regs, Uint64 LinearAddress, int Size, void *Data );
extern void	WriteLinear( tCPU *Regs, Uint64 LinearAddress, int Size, void *Data );
extern void	ReadSegment( tCPU *Regs, tSReg *Segment, Uint64 Address, int Size, void *Data );
extern void	WriteSegment( tCPU *Regs, tSReg *Segment, Uint64 Address, int Size, void *Data );
extern Uint64	ReadSegment64( tCPU *Regs, tSReg *Segment, Uint64 Address );
extern Uint32	ReadSegment32( tCPU *Regs, tSReg *Segment, Uint64 Address );
extern Uint16	ReadSegment16( tCPU *Regs, tSReg *Segment, Uint64 Address );
extern Uint8	ReadSegment8( tCPU *Regs, tSReg *Segment, Uint64 Address );
extern void	WriteSegment64( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint64 Value );
extern void	WriteSegment32( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint32 Value );
extern void	WriteSegment16( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint16 Value );
extern void	WriteSegment8( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint8 Value );
// --- Exceptions ---
extern void	PushInterrupt(tCPU *Regs);
extern int	Exception_GPF( tCPU *Regs, Uint16 ErrorCode );
extern int	Exception_UndefinedOpcode( tCPU *Regs );
extern int	Exception_PageFault(tCPU *Regs, Uint64 LinearAddress, Uint16 ErrorCode);

#endif
