/* EmulOS 2
 * x86/x86-64 CPU Driver
 * By thePowersGang
 * 
 * decode.c
 * - Opcode Decoding and Execution
 */
#include <emulos.h>
#include "common.h"
#include "opcodes.h"
#include "voodoo.h"

// === MACROS ===
#define TRACE(v...)	do{if(Regs->Debug.Trace)Debug(v);}while(0)

// === PROTOTYPES ===
Uint64	GetRegister64( tCPU *Regs, int Num );
Uint32	GetRegister32( tCPU *Regs, int Num );
Uint16	GetRegister16( tCPU *Regs, int Num );
Uint8	GetRegister8( tCPU *Regs, int Num );
void	SetRegister64( tCPU *Regs, int Num, Uint64 Value );
void	SetRegister32( tCPU *Regs, int Num, Uint32 Value );
void	SetRegister16( tCPU *Regs, int Num, Uint16 Value );
void	SetRegister8( tCPU *Regs, int Num, Uint8 Value );
Uint64	ProcessModRM(tCPU *Regs, int *IPOffset, int *Mod, int *Reg);
Uint32	DecodeSIB32(tCPU *Regs, int *IPOffset);
Uint64	DecodeSIB64(tCPU *Regs, int *IPOffset);

// === CONSTANTS ===
const char *cReg8Names[] = {"AL", "CL", "DL", "BL", "AH", "CH", "DH", "BH"};
const char *cReg8NamesR[] = {"AL", "CL", "DL", "BL", "SPb", "BPb", "SIb", "DIb"
	"R8b", "R9b", "R10b", "R11b", "R12b", "R13b", "R14b", "R15b"};
const char *cReg16Names[] = {"AX", "CX", "DX", "BX", "SP", "BP", "SI", "DI",
	"R8w", "R9w", "R10w", "R11w", "R12w", "R13w", "R14w", "R15w"};
const char *cReg32Names[] = {"EAX", "ECX", "EDX", "EBX", "ESP", "EBP", "ESI", "EDI",
	"R8d", "R9d", "R10d", "R11d", "R12d", "R13d", "R14d", "R15d"};
const char *cReg64Names[] = {"RAX", "RCX", "RDX", "RBX", "RSP", "RBP", "RSI", "RDI",
	"R8", "R9", "R10", "R11", "R12", "R13", "R14", "R15"};

// === CODE ===
/**
 * \fn int OpcodeDecode(tCPU *Regs)
 * \brief Decodes and executes an opcode
 */
int OpcodeDecode(tCPU *Regs)
{
	Uint8	opcode;
	
	 int	wasPrefix = 1;
	 int	ipOffset = 0;
	Uint64	initialIP = Regs->RIP;
	tSReg	initialCS = Regs->CS;
	 int	mod, reg;
	
	Uint8	val8, val8_2;
	Uint16	val16, val16_2;
	Uint32	val32, val32_2;
	Uint64	val64, val64_2;
	Uint64	addr;
	
	if(Regs->Debug.Trace)
	{
		Debug("(%2).%u ", Regs->Num, Regs->Timestamp);
		switch(Regs->Mode)
		{
		case MODE_RM:
		case MODE_VM8086:
			Debug("%4:%4 ", Regs->CS.Sel, (Uint16)(Regs->RIP&0xFFFF));
			break;
		case MODE_PM:
		case MODE_COMPAT:
			Debug("%4:%8 ", Regs->CS.Sel, (Uint32)(Regs->RIP&0xFFFFFFFF));
			break;
		case MODE_LM:
			Debug("%8:%8 ", Regs->RIP);
			break;
		}
	}
	
	// Reset decoding state
	Regs->NestedExceptions = 0;
	Regs->AtomicAccess = 0;
	Regs->OvrAddress = 0;
	Regs->OvrOperand = 0;
	Regs->ModRM.Mod = 0;	Regs->ModRM.Reg = 0;	Regs->ModRM.RM = 0;
	Regs->SIB.Scale = 0;	Regs->SIB.Index = 0;	Regs->SIB.Base = 0;
	Regs->REX.Valid = 0;
	Regs->Seg = NULL;	// Set in ProcessModRM
	
	// === Process Prefixes ===
	do {
		opcode = ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset++ );
		
		switch(opcode)
		{
		// Lock Memory Bus
		case PREF_LOCK:
			if( Regs->AtomicAccess ) {
				Memory_UnlockBus(Regs->AtomicAccess);
				return Exception_UndefinedOpcode(Regs);	// Should be GPF I think
			}
			Regs->AtomicAccess = Memory_LockBus();
			TRACE("LOCK ");
			break;
		
		case PREF_REPNZ:
			break;
		
		case PREF_REPZ:	// Also REP
			break;
		
		// --- Segment Overrides
		case PREF_CS:	break;
		case PREF_SS:	break;
		case PREF_DS:
			TRACE("DS: ");
			Regs->Seg = &Regs->DS;
			break;
		case PREF_ES:
			TRACE("ES: ");
			Regs->Seg = &Regs->ES;
			break;
		case PREF_FS:
			TRACE("FS: ");
			Regs->Seg = &Regs->FS;
			break;
		case PREF_GS:
			TRACE("GS: ");
			Regs->Seg = &Regs->GS;
			break;
		
		// --- Branch Hints (Ignored)
		// Note ONLY IN 64-bit mode
		//case PREF_NOTTAKEN:
		//case PREF_TAKEN:
		//	break;
		
		// --- Operand Size Override
		case PREF_OPRSIZE:
			break;
		// --- Address Size Override
		case PREF_ADDRSIZE:
			break;
		
		default:
			wasPrefix = 0;
			break;
		}
	} while(wasPrefix);
	
	
	//TRACE("{%2} ", opcode);
	
	// === Process Opcodes ===
	switch( opcode )
	{
	
	// ----------------
	// ALU (Arithmatic)
	// ----------------
	
	#define DEBUG_ALU(t)	do{switch((t)){\
		case 0:	TRACE("ADD");	break;\
		case 1:	TRACE("OR");	break;\
		case 4:	TRACE("AND");	break;\
		}}while(0)
	#define DO_ALU_FUNC(t, d, s, b)	do{switch((t)) {\
		case 0:	DO_ADD( (d), (s), (b) );	break;\
		case 1:	DO_OR( (d), (s), (b) );	break;\
		case 4:	DO_AND( (d), (s), (b) );	break;\
		default:TRACE("#UD ALU %i\n", t);return Exception_UndefinedOpcode(Regs);\
		}}while(0)
	
	// <op> r/m8, imm8
	case OP_ARITH_MI:
		addr = GETMODRM();
		DEBUG_ALU(reg);	TRACE(".MI ");
		// 8-bit Sign Extended to 64-bit
		if( USE_EXTENDTO64_OPERAND ) {
			val64_2 = (Sint8) ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset ++;
			GET_RM64();
			DO_ALU_FUNC( reg, val64, val64_2, 64 );
			SET_RM64();
		}
		// 8-bit
		else {
			val8_2 = ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset ++;
			GET_RM8();
			DO_ALU_FUNC( reg, val8, val8_2, 8 );
			SET_RM8();
		}
		break;
	
	// <op> r/m(16,32,64), imm(16,32)
	case OP_ARITH_MIX:
		addr = GETMODRM();
		DEBUG_ALU(reg);	TRACE(".MIX ");
		// 32-bit Sign Extended to 64-bit
		if( USE_EXTENDTO64_OPERAND ) {
			val64_2 = (Sint32) ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
			GET_RM64();
			DO_ALU_FUNC( reg, val64, val64_2, 64 );
			SET_RM64();
			TRACE("%X", val64_2);
		}
		else if( USE_32BIT_OPERAND ) {
			val32_2 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
			GET_RM32();
			DO_ALU_FUNC( reg, val32, val32_2, 32 );
			SET_RM32();
			TRACE("%8", val32_2);
		}
		else {
			val16_2 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 2;
			GET_RM16();
			DO_ALU_FUNC( reg, val16, val16_2, 16 );
			SET_RM16();
			TRACE("%4", val16_2);
		}
		break;
	
	// <op> r/m(16,32,64), imm8
	case OP_ARITH_MI8:
		addr = GETMODRM();
		DEBUG_ALU(reg);	TRACE(".MI8 ");
		if( USE_EXTENDTO64_OPERAND ) {
			val64_2 = (Sint8) ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 1;
			GET_RM64();
			DO_ALU_FUNC( reg, val64, val64_2, 64 );
			SET_RM64();
			TRACE("%X", val64_2);
		}
		else if( USE_32BIT_OPERAND ) {
			val32_2 = (Sint8) ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 1;
			GET_RM32();
			DO_ALU_FUNC( reg, val32, val32_2, 32 );
			SET_RM32();
			TRACE("%8", val32_2);
		}
		else {
			val16_2 = (Sint8) ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 1;
			GET_RM16();
			DO_ALU_FUNC( reg, val16, val16_2, 16 );
			SET_RM16();
			TRACE("%4", val16_2);
		}
		break;
	
	// <op> AL, Imm8
	case OP_ADD_AI:	// 0
	case OP_OR_AI:	// 0
	case OP_AND_AI:	// 4
		DEBUG_ALU( (opcode>>3)&7 );	TRACE(".AI ");
		// Mode Agnostic
		val8 = ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset++ );
		DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[REG_A].L, val8, 8 );
		break;
	
	// <op> ([RE])AX, Imm16/32
	case OP_ADD_AIX:
	case OP_OR_AIX:
	case OP_AND_AIX:
		DEBUG_ALU( (opcode>>3)&7 );	TRACE(".AIX ");
		// 32-bit Sign Extended to 64-bit
		if( USE_EXTENDTO64_OPERAND ) {
			val64 = (Sint32) ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
			DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[REG_A].R, val64, 64 );
		}
		// 32-bit
		else if( USE_32BIT_OPERAND ) {
			val32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
			DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[REG_A].E, val32, 32 );
		}
		// 16-bit
		else {
			val16 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 2;
			DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[REG_A].X, val16, 16 );
		}
		break;
	
	// <alu> r/m8, r8
	case OP_ADD_MR:
	case OP_OR_MR:
	case OP_AND_MR:
		addr = GETMODRM();
		DEBUG_ALU( (opcode>>3)&7 );	TRACE(".MR ");
		GET_RM8();
		// Long mode (16 registers)
		if( Regs->Mode == MODE_LM && Regs->REX.Valid ) {
			DO_ALU_FUNC( (opcode>>3)&7, val8, Regs->GPR[reg].L, 8 );
		}
		// Others (8 registers including access to ?H
		else {
			if( reg < 4 )
				DO_ALU_FUNC( (opcode>>3)&7, val8, Regs->GPR[reg].L, 8 );
			else
				DO_ALU_FUNC( (opcode>>3)&7, val8, Regs->GPR[reg-4].H, 8 );
		}
		SET_RM8();
		break;
	
	// <alu> r/m16, r16
	case OP_ADD_MRX:
	case OP_OR_MRX:
	case OP_AND_MRX:
		addr = GETMODRM();
		DEBUG_ALU( (opcode>>3)&7 );	TRACE(".MRX ");
		// 64-bits
		if( USE_EXTENDTO64_OPERAND ) {
			GET_RM64();
			DO_ALU_FUNC( (opcode>>3)&7, val64, Regs->GPR[reg].R, 64);
			SET_RM64();
		}
		// 32-bits
		else if( USE_32BIT_OPERAND ) {
			GET_RM32();
			DO_ALU_FUNC( (opcode>>3)&7, val32, Regs->GPR[reg].E, 32);
			SET_RM32();
		}
		// 16-bits
		else {
			GET_RM16();
			DO_ALU_FUNC( (opcode>>3)&7, val16, Regs->GPR[reg].X, 16);
			SET_RM16();
		}
		break;
	
	// <alu> r8, r/m8
	case OP_ADD_RM:
	case OP_OR_RM:
	case OP_AND_RM:
		addr = GETMODRM();
		DEBUG_ALU( (opcode>>3)&7 );	TRACE(".RM ");
		GET_RM8();
		// Long mode (16 registers)
		if( Regs->Mode == MODE_LM && Regs->REX.Valid ) {
			DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[reg].L, val8, 8 );
		}
		// Others (8 registers, including access to ?H)
		else {
			if( reg < 4 )
				DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[reg].L, val8, 8 );
			else
				DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[reg-4].H, val8, 8 );
		}
		SET_RM8();
		break;
	
	// <alu> r16, r/m16
	case OP_ADD_RMX:
	case OP_OR_RMX:
	case OP_AND_RMX:
		addr = GETMODRM();
		DEBUG_ALU( (opcode>>3)&7 );	TRACE(".RMX ");
		// 64-bits
		if( USE_EXTENDTO64_OPERAND ) {
			GET_RM64();
			DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[reg].R, val64, 64);
			SET_RM64();
		}
		// 32-bits
		else if( USE_32BIT_OPERAND ) {
			GET_RM32();
			DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[reg].E, val32, 32);
			SET_RM32();
		}
		// 16-bits
		else {
			GET_RM16();
			DO_ALU_FUNC( (opcode>>3)&7, Regs->GPR[reg].X, val16, 16);
			SET_RM16();
		}
		break;
	
	// dec r16
	CASE8(OP_DEC_R):
		reg = opcode & 7;
		if( USE_32BIT_OPERAND ) {
			TRACE("DEC.R %s", cReg32Names[reg]);
			Regs->GPR[reg].E --;
			// TODO: Set Flags
			Regs->RFlags.Z = (Regs->GPR[reg].E == 0);
			Regs->RFlags.S = (Regs->GPR[reg].E >> 31);
			Regs->RFlags.O = (Regs->GPR[reg].E == 0xFFFFFFFF);
			// AF and PF? Set but what to
		}
		else {
			TRACE("DEC.R %s", cReg16Names[reg]);
			Regs->GPR[reg].X --;
			// TODO: Set Flags
			Regs->RFlags.Z = (Regs->GPR[reg].X == 0);
			Regs->RFlags.S = (Regs->GPR[reg].X >> 15);
			Regs->RFlags.O = (Regs->GPR[reg].X == 0xFFFF);
			// AF and PF? Set but what to
		}
		break;
	
	// ===
	// Load and Store (MOV family)
	// ===
	case OP_MOV_MR:
		TRACE("MOV.MR ");
		addr = GETMODRM();
		GET_RM8();
		val8 = GetRegister8( Regs, reg );
		SET_RM8();
		break;
	
	// mov sreg, r/m16
	case OP_MOV_SM:
		TRACE("MOV.SM ");
		addr = GETMODRM();
		GET_RM16();
		switch(reg)
		{
		case 0:	// ES
			TRACE("ES ");
			SetSegment(Regs, &Regs->ES, val16);
			break;
		case 3:	// DS
			TRACE("DS ");
			SetSegment(Regs, &Regs->DS, val16);
			break;
		default:
			TRACE("S%1 ", reg);
			TRACE("---INCOMPLETE---");
			break;
		}
		break;
	
	// mov r8, imm8
	CASE8(OP_MOV_RI):
		TRACE("MOV.RI ");
		val8 = ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset );
		ipOffset ++;
		reg = opcode & 7;
		SetRegister8( Regs, opcode&7, val8 );	// TODO: Check if REX.R is valid/usable here
		TRACE("%2", val8);
		break;
	
	// mov r16, imm16
	CASE8(OP_MOV_RIX):
		reg = opcode & 7;
		TRACE("MOV.RIX ");
		if( USE_EXTENDTO64_OPERAND ) {
			TRACE("%s ", cReg64Names[reg]);
			val64 = ReadSegment64( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 8;
			Regs->GPR[reg].R = val64;	// TODO: Check if REX.R is valid/usable here
			TRACE("%X", val64);
		}
		else if( USE_32BIT_OPERAND ) {
			TRACE("%s ", cReg32Names[reg]);
			val32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
			Regs->GPR[reg].E = val32;	// TODO: Check if REX.R is valid/usable here
			TRACE("%8", val32);
		}
		else {
			TRACE("%s ", cReg16Names[reg]);
			val16 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 2;
			Regs->GPR[reg].X = val16;	// TODO: Check if REX.R is valid/usable here
			TRACE("%4", val16);
		}
		break;

	// mov r/m16, imm16
	case OP_MOV_MIX:
		TRACE("MOV.MIX ");
		addr = GETMODRM();
		if( USE_EXTENDTO64_OPERAND ) {
			GET_RM64();
			val64 = ReadSegment64( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 8;
			SET_RM64();
			TRACE("%X", val64);
		}
		else if( USE_32BIT_OPERAND ) {
			GET_RM32();
			val32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
			SET_RM32();
			TRACE("%8", val32);
		}
		else {
			GET_RM16();
			val16 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 2;
			SET_RM16();
			TRACE("%4", val16);
		}
		break;
	
	// ===
	// Execution Flow
	// ===
	case OP_JMP_FAR:	// jmp 16:(16,32)
		TRACE("JMP FAR ");
		
		// Get IP
		if( USE_32BIT_OPERAND ) {
			val32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
		}
		else {
			val32 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 2;
		}
		
		// Get CS
		val16 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + ipOffset );
		ipOffset += 2;
		TRACE("%4:", val16);
		if( USE_32BIT_OPERAND )
			TRACE("%8", val32);
		else
			TRACE("%4", val32);
		
		// Do the Jump
		SetSegment( Regs, &Regs->CS, val16 );
		Regs->RIP = val32 - ipOffset;
		break;
	case OP_JMP_REL:	// jmp rel8
		TRACE("JMP .");
		val8 = ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset++ );
		Regs->RIP += (Sint8)val8;
		if( (Sint8)val32 < 0 )
			TRACE("-%u", -(Sint8)val8);
		else
			TRACE("+%u", (Sint8)val8);
		break;
	case OP_JMP_RELX:	// jmp rel16/32
		TRACE("JMP .");
		if( USE_32BIT_OPERAND || USE_EXTENDTO64_OPERAND) {
			val32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 4;
			Regs->RIP += (Sint32)val32;
			if( (Sint32)val32 < 0 )
				TRACE("-0x%8", -(Sint32)val32);
			else
				TRACE("+0x%8", (Sint32)val32);
		}
		else {
			val16 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + ipOffset );
			ipOffset += 2;
			Regs->RIP += (Sint16)val16;
			if( (Sint16)val16 < 0 )
				TRACE("-0x%4", -(Sint16)val16);
			else
				TRACE("+0x%4", (Sint16)val16);
		}
		break;
	// Jcc rel8
	CASE16(OP_JCC_REL):
		val8 = ReadSegment8(Regs, &Regs->CS, Regs->RIP + ipOffset);
		ipOffset++;
		switch(opcode & 15)
		{
		case 0:	TRACE("JO ");	// JO
			if( Regs->RFlags.O )	Regs->RIP += (Sint8)val8;
			break;
		case 1:	TRACE("JNO ");	// JNO
			if( !Regs->RFlags.O )	Regs->RIP += (Sint8)val8;
			break;
		case 2:	TRACE("JC ");	// JC
			if( Regs->RFlags.C )	Regs->RIP += (Sint8)val8;
			break;
		case 3:	TRACE("JNC ");	// JNC
			if( !Regs->RFlags.C )	Regs->RIP += (Sint8)val8;
			break;
		case 4:	TRACE("JZ ");	// JZ
			if( Regs->RFlags.Z )	Regs->RIP += (Sint8)val8;
			break;
		case 5:	TRACE("JNZ ");	// JNZ
			if( !Regs->RFlags.Z )	Regs->RIP += (Sint8)val8;
			break;
		case 6:	TRACE("JNA ");	// JNA/JBE
			if( Regs->RFlags.C || Regs->RFlags.Z )	Regs->RIP += (Sint8)val8;
			break;
		case 7:	TRACE("JA ");	// JA/JNBE
			if( !Regs->RFlags.C && !Regs->RFlags.Z )	Regs->RIP += (Sint8)val8;
			break;
		default:
			TRACE("Jcc UNK %i ", opcode&15);
			break;
		}
		if( (Sint8)val8 < 0 )
			TRACE(".-%i", -(Sint8)val8);
		else
			TRACE(".+%i", (Sint8)val8);
		break;
	
	// - Halt CPU until interrupt
	case OP_HLT:
		TRACE("HLT");
		Regs->StateFlags.Halted = TRUE;
		break;
	
	// ===
	// Extended Opcodes
	// ===
	case 0x0F:
		opcode = ReadSegment8( Regs, &Regs->CS, Regs->RIP + ipOffset++ );
		switch( opcode )
		{
		case 0x01:
			addr = GETMODRM();
			switch(reg)
			{
			case 2:	TRACE("LGDT ");
				SetGDT( Regs, addr );
				break;
			case 3:	TRACE("LIDT ");
				SetIDT( Regs, addr );
				break;
			default:
				TRACE("#UD 0F 01 /%1\n", reg);
				Memory_UnlockBus(Regs->AtomicAccess);
				return Exception_UndefinedOpcode(Regs);
			}
			break;
		
		// 0x20 - Mov from CR
		case 0x20:	// mov r32, cr
			TRACE("MOV.RC ");
			addr = GETMODRM();
			GET_RM32();	// Value is not used
			switch( reg )
			{
			case 0:	TRACE("CR0 ");
				val64 = GetCR0(Regs);
				break;
			case 2:	TRACE("CR2 ");
				val64 = Regs->CR2;
				break;
			case 3:	TRACE("CR3 ");
				val64 = Regs->CR3;
				break;
			case 4:	TRACE("CR4 ");
				val64 = GetCR4(Regs);
				break;
			case 8:	TRACE("CR8 (UNIMPL) ");
				val64 = 0;
				break;
			default:
				TRACE("#UD 0F 20 /%1 (mov from CRs)\n", reg);
				Memory_UnlockBus(Regs->AtomicAccess);
				return Exception_UndefinedOpcode(Regs);
				break;
			}
			val32 = val64;
			
			if( USE_EXTENDTO64_OPERAND )
				SET_RM64();
			else
				SET_RM32();
			break;
		// 0x22 - Mov to CR
		case 0x22:	// mov cr, r32
			TRACE("MOV.RC ");
			addr = GETMODRM();
			TRACE("CR%i ", reg);
			if( USE_EXTENDTO64_OPERAND )
				GET_RM64();
			else {
				GET_RM32();
				val64 = val32;
			}
			switch( reg )
			{
			case 0:
				SetCR0(Regs, val64);
				break;
			case 2:
				Regs->CR2 = val64;
				break;
			case 3:
				Regs->CR3 = val64;
				break;
			case 4:
				SetCR4(Regs, val64);
				break;
			case 8:	TRACE("(UNIMPL) ");
				break;
			default:
				TRACE("#UD 0F 22 /%1 (mov to CRs)\n", reg);
				Memory_UnlockBus(Regs->AtomicAccess);
				return Exception_UndefinedOpcode(Regs);
				break;
			}			
			break;
		
		default:
			TRACE("#UD 0F %2\n", opcode);
			Memory_UnlockBus(Regs->AtomicAccess);
			return Exception_UndefinedOpcode(Regs);
		}
		break;
	
	case 0xFF:
		addr = GETMODRM();
		
		switch( reg )
		{
		// call r/m(16,32,64)
		case 2:
			if( Regs->Mode == MODE_LM ) {
				GET_RM64();
				Regs->RIP = val64 - ipOffset;
			}
			else if( USE_32BIT_OPERAND ) {
				GET_RM32();
				Regs->RIP = val32 - ipOffset;
			}
			else {
				GET_RM16();
				Regs->RIP = val16 - ipOffset;
			}
			break;
		
		default:
			TRACE("#UD FF %2\n", opcode);
			Memory_UnlockBus(Regs->AtomicAccess);
			Regs->AtomicAccess = 0;
			return Exception_UndefinedOpcode(Regs);
		}
		break;
	
	default:
		TRACE("#UD %2\n", opcode);
		Memory_UnlockBus(Regs->AtomicAccess);
		return Exception_UndefinedOpcode(Regs);
	}
	
	// Release the lock on the bus
	Memory_UnlockBus(Regs->AtomicAccess);
	Regs->AtomicAccess = 0;
	
	// Increment IP and wrap to current integer size
	Regs->RIP += ipOffset;
	switch(Regs->Mode)
	{
	case MODE_PM:	case MODE_COMPAT:
		Regs->RIP &= 0xFFFFFFFF;
		break;
	case MODE_RM:	case MODE_VM8086:
		Regs->RIP &= 0xFFFF;
		break;
	}
	
	Regs->Timestamp ++;
	
	if(Regs->Debug.Trace)
	{
		Debug("\t; ");
		while( ipOffset-- ) {
			Debug("%2", ReadSegment8(Regs, &initialCS, initialIP++));
		}
		Debug("\n");
	}
	// Return no-error
	return 0;
}

Uint64 GetRegister64(tCPU *Regs, int Num)
{
	return Regs->GPR[Num].R;
}

Uint32 GetRegister32(tCPU *Regs, int Num)
{
	return Regs->GPR[Num].E;
}

Uint16 GetRegister16( tCPU *Regs, int Num )
{
	return Regs->GPR[Num].X;
}
Uint8 GetRegister8( tCPU *Regs, int Num )
{
	if( Regs->Mode == MODE_LM && !Regs->REX.Valid ) {
		TRACE("%s", cReg8NamesR[Num]);
		return Regs->GPR[Num].L;
	}
	TRACE("%s", cReg8Names[Num]);
	// AL, CL, DL, BL
	if(Num < 4)
		return Regs->GPR[Num].L;
	// AH, CH, DH, BH
	return Regs->GPR[Num-4].H;
}

void SetRegister64( tCPU *Regs, int Num, Uint64 Value )
{
	Regs->GPR[Num].R = Value;
}
void SetRegister32( tCPU *Regs, int Num, Uint32 Value )
{
	Regs->GPR[Num].E = Value;
}
void SetRegister16( tCPU *Regs, int Num, Uint16 Value )
{
	Regs->GPR[Num].X = Value;
}
void SetRegister8( tCPU *Regs, int Num, Uint8 Value )
{
	if( Regs->Mode == MODE_LM && Regs->REX.Valid )
		Regs->GPR[Num].L = Value;
	// AL, CL, DL, BL
	if(Num < 4)
		Regs->GPR[Num].L = Value;
	// AH, CH, DH, BH
	Regs->GPR[Num-4].H = Value;
}

/**
 * \fn Uint64 ProcessModRM(tCPU *Regs, int *IPOffset, int *Mod, int *Reg)
 * \brief Reads the ModR/M Byte from the instruction stream and processes it
 * \return Virtual address OR a encoding of the register number
 */
Uint64 ProcessModRM(tCPU *Regs, int *IPOffset, int *Mod, int *Reg)
{
	Uint8	byte;
	Uint16	disp16;
	Uint32	disp32;
	
	// Read ModR/M Byte
	byte = ReadSegment8( Regs, &Regs->CS, Regs->RIP + *IPOffset );
	(*IPOffset) ++;
	
	Regs->ModRM.RM = byte&7;
	Regs->ModRM.Reg = (byte>>3)&7;
	Regs->ModRM.Mod = (byte>>6)&3;
	
	if(Mod)	*Mod = Regs->ModRM.Mod;
	if(Reg)	*Reg = Regs->ModRM.Reg;
	
	// 16-bit Mode
	if( Regs->Mode == MODE_RM || Regs->Mode == MODE_VM8086 )
	{
		
		switch(Regs->ModRM.Mod)
		{
		// No Displacement
		case 0:
			switch(Regs->ModRM.RM)
			{
			case 0:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:BX+SI] ", Regs->Seg->Name);
				return (Regs->GPR[REG_B].X + Regs->GPR[REG_SI].X) & 0xFFFF;
			case 1:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:BX+DI] ", Regs->Seg->Name);
				return (Regs->GPR[REG_B].X + Regs->GPR[REG_DI].X) & 0xFFFF;
			case 2:
				if(!Regs->Seg)	Regs->Seg = &Regs->SS;
				TRACE("[%s:BP+SI] ", Regs->Seg->Name);
				return (Regs->GPR[REG_BP].X + Regs->GPR[REG_SI].X) & 0xFFFF;
			case 3:
				if(!Regs->Seg)	Regs->Seg = &Regs->SS;
				TRACE("[%s:BP+DI] ", Regs->Seg->Name);
				return (Regs->GPR[REG_BP].X + Regs->GPR[REG_DI].X) & 0xFFFF;
			case 4:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:SI] ", Regs->Seg->Name);
				return Regs->GPR[REG_SI].X;
			case 5:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:DI] ", Regs->Seg->Name);
				return Regs->GPR[REG_DI].X;
			case 6:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				disp16 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + *IPOffset );
				TRACE("[%s:0x%4] ", Regs->Seg->Name, disp16);
				*IPOffset += 2;
				return disp16;
			case 7:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:BX] ", Regs->Seg->Name);
				return Regs->GPR[REG_B].X;
			}
		
		case 1:
			disp16 = (Sint8) ReadSegment8( Regs, &Regs->CS, Regs->RIP + *IPOffset );
			(*IPOffset) ++;
			goto comm16_1and2;
		
		case 2:
			disp16 = ReadSegment16( Regs, &Regs->CS, Regs->RIP + *IPOffset );
			(*IPOffset) ++;
		
		comm16_1and2:
			switch(Regs->ModRM.RM)
			{
			case 0:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:BX+SI+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_B].X + Regs->GPR[REG_SI].X + disp16) & 0xFFFF;
			case 1:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:BX+DI+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_B].X + Regs->GPR[REG_DI].X + disp16) & 0xFFFF;
			case 2:
				if(!Regs->Seg)	Regs->Seg = &Regs->SS;
				TRACE("[%s:BP+SI+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_BP].X + Regs->GPR[REG_SI].X + disp16) & 0xFFFF;
			case 3:
				if(!Regs->Seg)	Regs->Seg = &Regs->SS;
				TRACE("[%s:BP+DI+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_BP].X + Regs->GPR[REG_DI].X + disp16) & 0xFFFF;
			case 4:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:SI+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_SI].X + disp16) & 0xFFFF;
			case 5:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:DI+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_DI].X + disp16) & 0xFFFF;
			case 6:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:BP+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_BP].X + disp16) & 0xFFFF;
			case 7:
				if(!Regs->Seg)	Regs->Seg = &Regs->DS;
				TRACE("[%s:BX+0x%4] ", Regs->Seg->Name, disp16);
				return (Regs->GPR[REG_B].X + disp16) & 0xFFFF;
			}
			break;
		case 3:
			TRACE("%s ", cReg16Names[Regs->ModRM.RM]);
			return TO_REGADDR(Regs->ModRM.RM);
		}
		
		return 0;
	}
	
	if( Regs->REX.Valid )
	{
		Regs->ModRM.Reg |= (Regs->REX.R << 4);
		*Reg |= (Regs->REX.R << 4);
	}
	
	// 32-bit protected mode
	if( (Regs->Mode == MODE_PM || Regs->Mode == MODE_COMPAT)
	 || (Regs->Mode == MODE_LM && !Regs->REX.W) )
	{
		switch(Regs->ModRM.Mod)
		{
		// No Displacement
		case 0:
			if(!Regs->Seg)	Regs->Seg = &Regs->DS;
			if( Regs->ModRM.RM == 4 )
				return DecodeSIB32(Regs, IPOffset);
			if( Regs->ModRM.RM == 5 )
			{
				disp32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + *IPOffset );
				(*IPOffset) += 4;
				TRACE("[%s:%8] ", Regs->Seg->Name, disp32);
				return disp32;
			}
			TRACE("[%s:%s] ", Regs->Seg->Name, cReg32Names[Regs->ModRM.RM]);
			return Regs->GPR[Regs->ModRM.RM].E;
		
		// Sign Extended 8-bit displacement
		case 1:
			disp32 = (Sint8) ReadSegment8( Regs, &Regs->CS, Regs->RIP + *IPOffset );
			(*IPOffset) += 1;
			goto comm32_1and2;
		// 32-bit displacement
		case 2:
			disp32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + *IPOffset );
			(*IPOffset) += 4;
			
		comm32_1and2:
			if(!Regs->Seg)
				Regs->Seg = (Regs->ModRM.RM == 5) ? &Regs->SS : &Regs->DS;
			
			if( Regs->ModRM.RM == 4 ) {
				Uint64	ret = (DecodeSIB32(Regs, IPOffset) + disp32) & 0xFFFFFFFF;
				TRACE("%8] ",disp32);
				return ret;
			}
			
			TRACE("[%s:%s+%8] ", Regs->Seg->Name, cReg32Names[Regs->ModRM.RM], disp32);
			return (Regs->GPR[Regs->ModRM.RM].E + disp32) & 0xFFFFFFFF;
		
		case 3:
			TRACE("%s ", cReg32Names[Regs->ModRM.RM]);
			return TO_REGADDR(Regs->ModRM.RM);
		}
		return 0;
	}
	
	// 64-bit Mode
	switch(Regs->ModRM.Mod)
	{
	// No Displacement
	case 0:
		if(!Regs->Seg)	Regs->Seg = &Regs->DS;
		if( Regs->ModRM.RM == 4 || Regs->ModRM.RM == 12 )
			return DecodeSIB64(Regs, IPOffset);
		if( Regs->ModRM.RM == 5 )
		{
			disp32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + *IPOffset );
			(*IPOffset) += 4;
			TRACE("[%s:%8] ", Regs->Seg->Name, disp32);
			return disp32;
		}
		TRACE("[%s:%s] ", Regs->Seg->Name, cReg32Names[Regs->ModRM.RM]);
		return Regs->GPR[Regs->ModRM.RM].R;
	
	// Sign Extended 8-bit displacement
	case 1:
		disp32 = (Sint8) ReadSegment8( Regs, &Regs->CS, Regs->RIP + *IPOffset );
		(*IPOffset) += 1;
		goto comm64_1and2;
	// 32-bit displacement
	case 2:
		disp32 = ReadSegment32( Regs, &Regs->CS, Regs->RIP + *IPOffset );
		(*IPOffset) += 4;
		
	comm64_1and2:
		if(!Regs->Seg)
			Regs->Seg = (Regs->ModRM.RM == 5) ? &Regs->SS : &Regs->DS;
		
		if( Regs->ModRM.RM == 4 )
			return DecodeSIB64(Regs, IPOffset) + (Sint32)disp32;
		
		return Regs->GPR[Regs->ModRM.RM].R + (Sint32)disp32;
	
	case 3:
		return TO_REGADDR(Regs->ModRM.RM);
	}
	return 0;
}

/**
 * \fn Uint32 DecodeSIB32(tCPU *Regs, int *IPOffset)
 * \brief Decodes the SIB byte and returns the resultant address
 */
Uint32 DecodeSIB32(tCPU *Regs, int *IPOffset)
{
	Uint8	byte;
	Uint32	base, index;
	
	byte = ReadSegment8( Regs, &Regs->CS, Regs->RIP + *IPOffset );
	*IPOffset += 1;
	
	Regs->SIB.Scale = (byte>>6)&3;
	Regs->SIB.Index = (byte>>3)&7;
	Regs->SIB.Base = (byte>>0)&7;
	
	if(Regs->REX.Valid)
	{
		Regs->SIB.Index |= (Regs->REX.X << 4);
		Regs->SIB.Base |= (Regs->REX.B << 4);
	}
	
	TRACE("[");
	
	if( Regs->SIB.Base == REG_BP && Regs->ModRM.Mod == 0 )
		base = 0;
	else {
		TRACE("%s+", cReg32Names[Regs->SIB.Base]);
		base = Regs->GPR[ Regs->SIB.Base ].E;
	}
	
	if( Regs->SIB.Index == REG_SP )
		index = 0;
	else {
		TRACE("%s*%i+", cReg32Names[Regs->SIB.Index], 1 << Regs->SIB.Scale);
		index = Regs->GPR[ Regs->SIB.Index ].E << Regs->SIB.Scale;
	}
	
	return base + index;
}

/**
 * \fn Uint64 DecodeSIB64(tCPU *Regs, int *IPOffset)
 * \brief Decodes the SIB byte and returns the resultant address
 */
Uint64 DecodeSIB64(tCPU *Regs, int *IPOffset)
{
	Uint8	byte;
	Uint32	base, index;
	
	byte = ReadSegment8( Regs, &Regs->CS, Regs->RIP + *IPOffset );
	*IPOffset += 1;
	
	Regs->SIB.Scale = (byte>>6)&3;
	Regs->SIB.Index = (byte>>3)&7;
	Regs->SIB.Base = (byte>>0)&7;
	
	if(Regs->REX.Valid)
	{
		Regs->SIB.Index |= (Regs->REX.X << 4);
		Regs->SIB.Base |= (Regs->REX.B << 4);
	}
	
	if( Regs->SIB.Base == REG_BP && Regs->ModRM.Mod == 0 )
		base = 0;
	else
		base = Regs->GPR[ Regs->SIB.Base ].R;
	
	if( Regs->SIB.Index == REG_SP )
		index = 0;
	else
		index = Regs->GPR[ Regs->SIB.Index ].R << Regs->SIB.Scale;
	
	return base + index;
}
