/* EmulOS 2
 * x86/x86-64 CPU Driver
 * By thePowersGang
 * 
 * descs.c
 * - Descriptors
 */
#include <emulos.h>
#include "common.h"

// === PROTOTYPES ===
void	SetSegment( tCPU *Regs, tSReg *Seg, Uint16 Selector );
void	SetGDT( tCPU *Regs, Uint64 Addr );
void	SetIDT( tCPU *Regs, Uint64 Addr );

// === CODE ===
/**
 * \fn void SetSegment( tCPU *Regs, tSReg *Seg, Uint16 Selector )
 * \brief Sets a segment register
 */
void SetSegment( tCPU *Regs, tSReg *Seg, Uint16 Selector )
{
	Uint64	dtPtr;
	Uint32	dtent[4];
	 int	dpl;
	
	// Handle mode changes
	if( Regs->Mode == MODE_PM && !Regs->CR0.PE && Seg == &Regs->CS ) {
		Regs->Mode = MODE_RM;
	}
	else if( Regs->Mode == MODE_RM && Regs->CR0.PE && Seg == &Regs->CS ) {
		Regs->Mode = MODE_PM;
	}
		//Debug("-- ENTERED PROTECTED MODE --");
	
	switch(Regs->Mode)
	{
	case MODE_RM:
	case MODE_VM8086:
		Seg->Sel = Selector;
		Seg->Base = (Uint32)Selector * 16;
		Debug("%s = %4", Seg->Name, Selector);
		break;
	
	case MODE_PM:
		if( Selector & 4 ) {
			if( (Selector | 7) > Regs->LDTR.Limit ) {
				Exception_GPF( Regs, Selector );
				return;
			}
			dtPtr = Regs->LDTR.Base + (Selector & ~7);
		}
		else {
			if( (Selector | 7) > Regs->GDTR.Limit ) {
				Exception_GPF( Regs, Selector );
				return;
			}
			dtPtr = Regs->GDTR.Base + (Selector & ~7);
		}
		ReadLinear( Regs, dtPtr, 8, &dtent );
		
		// Check Valid Bit
		if( !(dtent[1] & 0x8000) ) {
			Exception_GPF( Regs, Selector );
			return;
		}
		
		// Base
		Seg->Base  = (dtent[0] >> 16) & 0xFFFF;
		Seg->Base |= (dtent[1] & 0xFF) << 16;
		Seg->Base |= dtent[1] & 0xFF000000;
		
		// Limit
		Seg->Limit  = dtent[0] & 0xFFFF;
		Seg->Limit |= dtent[1] & 0xF0000;
		if(dtent[1] & 0x800000)
		{
			Seg->Limit *= 0x1000;
			Seg->Limit |= 0xFFF;
		}
		
		// Flags
		Seg->Flags.Code = (dtent[1] & 0x800) ? TRUE : FALSE;
		Seg->Flags.DC = (dtent[1] & 0x400) ? TRUE : FALSE;
		Seg->Flags.RW = (dtent[1] & 0x200) ? TRUE : FALSE;
		Seg->Flags.Size = (dtent[1] & 0x400000) ? TRUE : FALSE;
		
		dpl = (dtent[1] >> 13) & 3;
		if( dpl != (Selector & 3) ) {
			Exception_GPF( Regs, Selector );
			return;
		}
		
		if( Seg->Flags.Code && Seg->Flags.DC ) {
				if( dpl > Regs->RFlags.IOPL ) {
				Exception_GPF( Regs, Selector );
				return;
			}
		}
		else {
			if( dpl != Regs->RFlags.IOPL ) {
				Exception_GPF( Regs, Selector );
				return;
			}
		}
		Seg->Flags.DPL = Selector & 3;
		
		Seg->Sel = Selector;
		break;
	
	default:
		Debug("SetSegment: TODO (%i)\n", Regs->Mode);
		break;
	}
}

/**
 * \fn void SetGDT( tCPU *Regs, Uint64 Addr )
 */
void SetGDT( tCPU *Regs, Uint64 Addr )
{
	tDesc	new = {0};
	
	// Only PL0 Can change the GDT
	if( Regs->RFlags.IOPL != 0 ) {
		Exception_GPF(Regs, 0);
		return ;
	}
	
	if( Regs->Mode == MODE_COMPAT || Regs->Mode == MODE_LM )
	{
		ReadSegment( Regs, Regs->Seg, Addr, 2+8, &new );
		Regs->GDTR.Limit = new.Limit;
		Regs->GDTR.Base = new.Base;
	}
	
	ReadSegment( Regs, Regs->Seg, Addr, 2+4, &new );
	Regs->GDTR.Limit = new.Limit;
	if( Regs->Mode == MODE_RM )
		Regs->GDTR.Base = new.Base & 0xFFFFFF;
	else
		Regs->GDTR.Base = new.Base & 0xFFFFFFFF;
}

/**
 * \fn void SetIDT( tCPU *Regs, Uint64 Addr )
 */
void SetIDT( tCPU *Regs, Uint64 Addr )
{
	tDesc	new = {0};
	
	// Only PL0 Can change the GDT
	if( Regs->RFlags.IOPL != 0 ) {
		Exception_GPF(Regs, 0);
		return ;
	}
	
	if( Regs->Mode == MODE_COMPAT || Regs->Mode == MODE_LM )
	{
		ReadSegment( Regs, Regs->Seg, Addr, 2+8, &new );
		Regs->IDTR.Limit = new.Limit;
		Regs->IDTR.Base = new.Base;
	}
	
	ReadSegment( Regs, Regs->Seg, Addr, 2+4, &new );
	Regs->IDTR.Limit = new.Limit;
	if( Regs->Mode == MODE_RM )
		Regs->IDTR.Base = new.Base & 0xFFFFFF;
	else
		Regs->IDTR.Base = new.Base & 0xFFFFFFFF;
}
