/* EmulOS 2
 * x86/x86-64 CPU Driver
 * By thePowersGang
 * 
 * memory.c
 * - Memory and Segment Handling
 */
#include <emulos.h>
#include "common.h"

// === PROTOTYPES ===
Uint8	ReadSegment8( tCPU *Regs, tSReg *Segment, Uint64 Address );
Uint16	ReadSegment16( tCPU *Regs, tSReg *Segment, Uint64 Address );
Uint32	ReadSegment32( tCPU *Regs, tSReg *Segment, Uint64 Address );
Uint64	ReadSegment64( tCPU *Regs, tSReg *Segment, Uint64 Address );

void	WriteSegment64( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint64 Value );
void	WriteSegment32( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint32 Value );
void	WriteSegment16( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint16 Value );
void	WriteSegment8( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint8 Value );

void	ReadLinear(tCPU *Regs, Uint64 LinearAddress, int Size, void *Data);
void	WriteLinear(tCPU *Regs, Uint64 LinearAddress, int Size, void *Data);

// === CODE ===
// --- Segments ---
void ReadSegment( tCPU *Regs, tSReg *Segment, Uint64 Address, int Size, void *Data )
{
	Uint64	laddr;
	laddr = Segment->Base + (Address & Segment->Limit);
	ReadLinear( Regs, laddr, Size, Data );
}
void WriteSegment( tCPU *Regs, tSReg *Segment, Uint64 Address, int Size, void *Data )
{
	Uint64	laddr;
	laddr = Segment->Base + (Address & Segment->Limit);
	WriteLinear( Regs, laddr, Size, Data );
}

Uint8 ReadSegment8( tCPU *Regs, tSReg *Segment, Uint64 Address )
{
	Uint8	data;
	ReadSegment( Regs, Segment, Address, 1, &data );
	return data;
}
Uint16 ReadSegment16( tCPU *Regs, tSReg *Segment, Uint64 Address )
{
	Uint16	data;
	ReadSegment( Regs, Segment, Address, 2, &data );
	return data;
}
Uint32 ReadSegment32( tCPU *Regs, tSReg *Segment, Uint64 Address )
{
	Uint32	data;
	ReadSegment( Regs, Segment, Address, 4, &data );
	return data;
}
Uint64 ReadSegment64( tCPU *Regs, tSReg *Segment, Uint64 Address )
{
	Uint64	data;
	ReadSegment( Regs, Segment, Address, 8, &data );
	return data;
}

void WriteSegment8( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint8 Value )
{
	WriteSegment( Regs, Segment, Address, 1, &Value );
}
void WriteSegment16( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint16 Value )
{
	WriteSegment( Regs, Segment, Address, 2, &Value );
}
void WriteSegment32( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint32 Value )
{
	WriteSegment( Regs, Segment, Address, 4, &Value );
}
void WriteSegment64( tCPU *Regs, tSReg *Segment, Uint64 Address, Uint64 Value )
{
	WriteSegment( Regs, Segment, Address, 8, &Value );
}

// --- Linear Memory (Paging) ---
Uint64 GetPhysAddr( tCPU *Regs, Uint64 LinearAddress )
{
	// Is paging enabled? (Plus, real mode doesn't use paging)
	if( !Regs->CR0.PG || Regs->Mode == MODE_RM )
		return LinearAddress;
	
	if( Regs->Mode == MODE_LM || Regs->Mode == MODE_COMPAT )
	{
		// - 64-bit mode paging!
		return 0;
	}
	
	// Ok, just 32-bit mode :(
	
	if( Regs->CR4.PAE )
	{
		// But with PAE! Fun time.
		return 0;
	}
	
	// el-plaino 32-bit mode paging
	{
		Uint32	dirent = 0, tabent = 0;
		
		Memory_Read(Regs->AtomicAccess, Regs->CR3+(LinearAddress>>22)*4, 4, &dirent);
		
		if( !(dirent & 1) ) {
			Exception_PageFault(Regs, LinearAddress, 0);
			return -1;
		}
		
		Memory_Read(Regs->AtomicAccess, (dirent&~0xFFF)+(LinearAddress>>12)*4, 4, &tabent);
		
		if( !(tabent & 1) ) {
			Exception_PageFault(Regs, LinearAddress, 0);
			return -1;
		}
	}
	
	return 0;
}

void ReadLinear(tCPU *Regs, Uint64 LinearAddress, int Size, void *Data)
{
	Uint64	paddr;
	if( LinearAddress & (Size-1) ) {
		// Non-Aligned
		while( Size -- )
		{
			paddr = GetPhysAddr( Regs, LinearAddress );
			Memory_Read(Regs->AtomicAccess, paddr, 1, Data);
			Data = (void*)((Uint)Data + 1);
			LinearAddress ++;
		}
	}
	else {
		// Aligned
		paddr = GetPhysAddr( Regs, LinearAddress );
		//Debug("Memory_Read(Atomic=%i, paddr=%X, Size=%i, Data=%8);\n",
		//	Atomic, paddr, Size, Data);
		Memory_Read(Regs->AtomicAccess, paddr, Size, Data);
	}
}

void WriteLinear(tCPU *Regs, Uint64 LinearAddress, int Size, void *Data)
{
	Uint64	paddr;
	if( (Size != 1 && Size != 2 && Size != 4 && Size != 8 ) || LinearAddress & (Size-1) ) {
		// Non-Aligned
		while( Size -- )
		{
			paddr = GetPhysAddr( Regs, LinearAddress );
			Memory_Write(Regs->AtomicAccess, paddr, 1, Data);
			Data = (void*)((Uint)Data + 1);
			LinearAddress ++;
		}
	}
	else {
		// Aligned
		paddr = GetPhysAddr( Regs, LinearAddress );
		Memory_Write(Regs->AtomicAccess, paddr, Size, Data);
	}
}
